import datetime as dt
from typing import Any, Union

from emitime.conversion import Number as Number
from emitime.conversion import add_conversion_methods as add_conversion_methods
from emitime.conversion import is_moment_str as is_moment_str
from emitime.conversion import is_time_str as is_time_str

LikeInterval = Union[str, dt.timedelta, dt.time, Interval]
LikeMoment = Union[str, dt.date, dt.datetime, Moment]

def upI(value: LikeInterval) -> dt.timedelta: ...
def upM(value: LikeMoment) -> dt.datetime: ...
def is_interval(value: Any) -> bool: ...
def is_moment(value: Any) -> bool: ...

class Interval:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def timedelta(self) -> dt.timedelta: ...
    @timedelta.setter
    def timedelta(self, value: LikeInterval) -> None: ...
    def __add__(
        self, other: Union["Interval", "Moment"]
    ) -> Union["Interval", "Moment"]: ...
    def __radd__(  # type: ignore
        self, other: Union["Interval", "Moment"]
    ) -> Union["Interval", "Moment"]: ...
    def __sub__(self, other: LikeInterval) -> Interval: ...
    def __rsub__(
        self, other: Union["Interval", "Moment"]
    ) -> Union["Interval", "Moment"]: ...
    def __mul__(self, other: Number) -> Interval: ...
    def __rmul__(self, other: Number) -> Interval: ...
    def __floordiv__(self, other: LikeInterval) -> int: ...
    def __mod__(self, other: LikeInterval) -> Interval: ...
    def __lt__(self, other: LikeInterval) -> bool: ...
    def __le__(self, other: LikeInterval) -> bool: ...
    def __eq__(self, other: LikeInterval) -> bool: ...  # type: ignore
    def __gt__(self, other: LikeInterval) -> bool: ...
    def __ge__(self, other: LikeInterval) -> bool: ...
    def __float__(self) -> float: ...

class Moment:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def datetime(self) -> dt.datetime: ...
    @datetime.setter
    def datetime(self, value: LikeMoment) -> None: ...
    @property
    def date(self) -> dt.date: ...
    @date.setter
    def date(self, value: dt.date) -> None: ...
    @date.setter
    def date(self, value: str) -> None: ...
    @property
    def time(self) -> dt.time: ...
    @time.setter
    def time(self, value: dt.time) -> None: ...
    @time.setter
    def time(self, value: str) -> None: ...
    def __rsub__(self, other: LikeMoment) -> Interval: ...
    def __add__(self, other: LikeInterval) -> Moment: ...
    def __lt__(self, other: LikeMoment) -> bool: ...
    def __le__(self, other: LikeMoment) -> bool: ...
    def __eq__(self, other: LikeMoment) -> bool: ...  # type: ignore
    def __gt__(self, other: LikeMoment) -> bool: ...
    def __ge__(self, other: LikeMoment) -> bool: ...
    def __float__(self) -> float: ...



__all__ = (
    "str_to_date",
    "date_to_str",
    "time_to_timedelta",
    "timedelta_to_str",
    "datetime_to_str",
    "str_to_timedelta",
    "timedelta_to_time",
    "str_to_time",
    "date_to_datetime",
    "str_to_datetime"
)

import datetime as dt
from collections import namedtuple
from typing import Union

from plum import add_conversion_method

Number = Union[int, float]

SplitSeconds = namedtuple(
    "SplitSeconds",
    ["d", "h", "m", "s", "ms", "us"],
    defaults=(0, 0, 0, 0, 0, 0),
)


def split_seconds(value: Number) -> SplitSeconds:
    if value < 0:
        msg = "the number of seconds must be greater than zero"
        raise ValueError(msg)
    msus = int(value * 1_000_000) % 1_000_000
    return SplitSeconds(
        d=int(value // 86_400),
        h=int(value % 86_400 // 3_600),
        m=int(value % 3_600 // 60),
        s=int(value % 60),
        ms=msus // 1_000,
        us=msus % 1_000,
    )


def microseconds(value: SplitSeconds) -> int:
    return (
        (86_400 * value.d + 3_600 * value.h + 60 * value.m + value.s) * 1_000_000
        + value.ms * 1_000
        + value.us
    )


def str_to_date(value: str) -> dt.date:
    if "-" in value:  # eng format
        ymd = value.split("-")
    elif "." in value:  # ru format
        ymd = value.split(".")[::-1]
    elif "/" in value:  # eng format
        ymd = value.split("/")
    else:
        raise NotImplementedError
    assert all([*map(str.isnumeric, ymd)])
    assert len(ymd) == 3
    year, month, day = ymd
    if len(year) == 2:
        # short year
        assert dt.date.today().year < 2035
        if int(year) <= 35:
            year = "20" + year
        else:
            year = "19" + year
    return dt.date(year=int(year), month=int(month), day=int(day))


def date_to_str(value: dt.date) -> str:
    y = value.year
    m = value.month
    d = value.day
    return f"{y}-{m:02}-{d:02}"


def time_to_timedelta(value: dt.time) -> dt.timedelta:
    secs = SplitSeconds(
        d=0, h=value.hour, m=value.minute, s=value.second, ms=0, us=value.microsecond
    )
    return dt.timedelta(microseconds=microseconds(secs))


def timedelta_to_str(value: dt.timedelta) -> str:
    sec = value.total_seconds()
    sign = "-" if sec < 0 else "+"
    s = split_seconds(abs(sec))
    time = f"{s.h:02}:{s.m:02}"
    if s.d != 0:
        time = f"{s.d}^{time}"
    tail = ""
    if s.us != 0:
        tail = f"'{s.us:03}"
    if s.ms != 0 or tail:
        tail = f".{s.ms:03}{tail}"
    if s.s != 0 or tail:
        tail = f":{s.s:02}{tail}"
    return f"{sign}{time}{tail}"


def time_to_str(value: dt.time) -> str:
    return timedelta_to_str(time_to_timedelta(value))


def datetime_to_str(value: dt.datetime) -> str:
    date = date_to_str(value.date())
    t = value.time()
    if t == dt.time():
        return date
    time = time_to_str(t)
    return f"{date}d{time}".replace("+", "")


def str_to_timedelta(value: str) -> dt.timedelta:
    try:
        sign = 1.0
        d = h = m = s = ms = us = 0
        if value.startswith("-"):
            sign = -1.0
            value = value[1:]
        if "^" in value or "d" in value:
            if "^" in value:
                split = value.split("^")
            else:
                split = value.split("d")
            value = split[-1]
            d = int(split[0])
        if value:
            if "." in value:
                split = value.split(".")
                value = split[0]
                msus = split[-1]
                if "'" in msus:
                    ms, us = [*map(int, msus.split("'"))]
                else:
                    ms = int(msus)
            split = value.split(":")
            h = int(split[0])
            m = int(split[1])
            if len(split) == 3:
                s = int(split[2])
        assert (
            0 <= h < 24
            and 0 <= m < 60
            and 0 <= s < 60
            and 0 <= ms < 1e3
            and 0 <= us < 1e3
        )
    except Exception:
        need_format = "Value does't match the format: [-|+][d^]hh:mm[:ss[.ms['us]]]"
        raise ValueError(need_format)
    secs = SplitSeconds(d=d, h=h, m=m, s=s, ms=ms, us=us)
    return sign * dt.timedelta(microseconds=microseconds(secs))


def timedelta_to_time(value: dt.timedelta) -> dt.time:
    total = value.total_seconds()
    if total < 0:
        msg = f"Negative '{value=!r}' can't be converted to time"
        raise ValueError(msg)
    secs = split_seconds(total)
    if secs.d != 0:
        msg = f"'{value=!r}' can't be converted to time as it contains days"
        raise ValueError(msg)

    return dt.time(
        hour=secs.h, minute=secs.m, second=secs.s, microsecond=secs.ms * 1_000 + secs.us
    )


def str_to_time(value: str) -> dt.time:
    try:
        assert "^" not in value
        assert "d" not in value
        assert "-" not in value
        assert "+" not in value
        timedelta = str_to_timedelta(value)
    except Exception:
        need_format = "'value' does not match the format: hh:mm[:ss[.ms['us]]]"
        raise ValueError(need_format)
    return timedelta_to_time(timedelta)


def date_to_datetime(value: dt.date) -> dt.datetime:
    if isinstance(value, dt.datetime):
        return value
    return dt.datetime(value.year, value.month, value.day)


def str_to_datetime(value: str) -> dt.datetime:
    """yyyy-mm-dd[^hh:mm[:ss[.ms['us]]]]"""
    try:
        if "^" in value or "d" in value:
            if "^" in value:
                split = value.split("^")
            else:
                split = value.split("d")
            assert len(split) == 2
            date = str_to_date(split[0])
            time = str_to_time(split[1])
        else:
            date = str_to_date(value)
            time = dt.time()

    except Exception:
        need_format = (
            "'value' does't match the format: yyyy-mm-dd[^hh:mm[:ss[.ms['us]]]]"
        )
        raise ValueError(need_format)

    return date_to_datetime(date) + time_to_timedelta(time)


def is_time_str(value: str) -> bool:
    try:
        str_to_timedelta(value)
    except Exception:
        return False
    return True


def is_moment_str(value: str) -> bool:
    if len(value) < 8:
        return False
    try:
        str_to_datetime(value)
    except Exception:
        return False
    return True


def add_conversion_methods():
    from emitime.base import Interval, Moment

    add_conversion_method(type_from=str, type_to=dt.timedelta, f=str_to_timedelta)
    add_conversion_method(type_from=dt.time, type_to=dt.timedelta, f=time_to_timedelta)
    add_conversion_method(
        type_from=Interval, type_to=dt.timedelta, f=lambda x: x.timedelta
    )

    add_conversion_method(type_from=str, type_to=dt.datetime, f=str_to_datetime)
    add_conversion_method(type_from=dt.date, type_to=dt.datetime, f=date_to_datetime)
    add_conversion_method(type_from=Moment, type_to=dt.datetime, f=lambda x: x.datetime)

    add_conversion_method(type_from=dt.datetime, type_to=str, f=datetime_to_str)
    add_conversion_method(type_from=dt.timedelta, type_to=str, f=timedelta_to_str)



import datetime as dt
import time
from typing import Any

from plum import convert, dispatch
from plum.type import PromisedType, Union

from emitime.conversion import (
    Number,
    add_conversion_methods,
    is_moment_str,
    is_time_str,
)

IntervalType = PromisedType()
MomentType = PromisedType()

# TODO: add float support
LikeInterval = Union[str, dt.timedelta, dt.time, IntervalType]
LikeMoment = Union[str, dt.date, dt.datetime, MomentType]


def upI(value: LikeInterval) -> dt.timedelta:
    return convert(value, dt.timedelta)


def upM(value: LikeMoment) -> dt.datetime:
    return convert(value, dt.datetime)


def is_interval(value: Any) -> bool:
    if isinstance(value, (dt.timedelta, dt.time, Interval)):
        return True
    if isinstance(value, str):
        if is_time_str(value):
            return True
    return False


def is_moment(value: Any) -> bool:
    if isinstance(value, (dt.date, dt.datetime, Moment)):
        return True
    if isinstance(value, str):
        if is_moment_str(value):
            return True
    return False


class Interval:
    def __init__(self, *args, **kwargs) -> None:
        if kwargs:
            raise NotImplementedError
        if len(args) != 1:
            raise NotImplementedError
        self.timedelta = args[0]

    @property
    def timedelta(self) -> dt.timedelta:
        return self._value

    @timedelta.setter
    @dispatch
    def timedelta(self, value: LikeInterval) -> None:
        self._value = upI(value)

    def __add__(
        self, other: Union["Interval", "Moment"]
    ) -> Union["Interval", "Moment"]:
        """this + (interval|moment) -> (interval|moment)"""
        if is_interval(other):
            return Interval(self.timedelta + upI(other))
        elif is_moment(other):
            return Moment(upM(other) + self.timedelta)
        raise NotImplementedError

    def __radd__(
        self, other: Union["Interval", "Moment"]
    ) -> Union["Interval", "Moment"]:
        """(interval|moment) + this -> (interval|moment)"""
        return self + other

    @dispatch
    def __sub__(self, other: LikeInterval) -> "Interval":
        """this - interval -> interval"""
        return Interval(self.timedelta - upI(other))

    def __rsub__(
        self, other: Union["Interval", "Moment"]
    ) -> Union["Interval", "Moment"]:
        """(interval|moment) - this -> (interval|moment)"""
        if is_interval(other):
            return Interval(upI(other) - self.timedelta)
        elif is_moment(other):
            return Moment(upM(other) - self.timedelta)
        raise NotImplementedError

    @dispatch
    def __mul__(self, other: Number) -> "Interval":
        """this * Number -> interval"""
        return Interval(self.timedelta * other)

    @dispatch
    def __rmul__(self, other: Number) -> "Interval":
        """Number * this -> interval"""
        return self * other

    @dispatch
    def __truediv__(self, other: LikeInterval) -> float:
        """this / interval -> float"""
        return self.timedelta / upI(other)

    @dispatch
    def __truediv__(self, other: Number) -> "Interval":
        """this / Number -> interval"""
        return Interval(self.timedelta / other)

    @dispatch
    def __floordiv__(self, other: LikeInterval) -> int:
        """this // interval -> int"""
        return self.timedelta // upI(other)

    @dispatch
    def __mod__(self, other: LikeInterval) -> "Interval":
        """this % interval -> interval"""
        return Interval(self.timedelta % upI(other))

    @dispatch
    def __lt__(self, other: LikeInterval) -> bool:
        """this < interval -> bool"""
        return self.timedelta < upI(other)

    @dispatch
    def __le__(self, other: LikeInterval) -> bool:
        """this <= interval -> bool"""
        return self.timedelta <= upI(other)

    @dispatch
    def __eq__(self, other: LikeInterval) -> bool:
        """this == interval -> bool"""
        return self.timedelta == upI(other)

    @dispatch
    def __gt__(self, other: LikeInterval) -> bool:
        """this > interval -> bool"""
        return self.timedelta > upI(other)

    @dispatch
    def __ge__(self, other: LikeInterval) -> bool:
        """this >= interval -> bool"""
        return self.timedelta >= upI(other)

    def __str__(self) -> str:
        return convert(self.timedelta, str)

    def __repr__(self) -> str:
        return str(self)

    def __float__(self) -> float:
        raise NotImplementedError


class Moment:
    def __init__(self, *args, **kwargs) -> None:
        if kwargs:
            raise NotImplementedError
        if len(args) != 1:
            raise NotImplementedError
        self.datetime = args[0]

    @property
    def datetime(self) -> dt.datetime:
        return self._value

    @datetime.setter
    @dispatch
    def datetime(self, value: LikeMoment) -> None:
        self._value = upM(value)

    @property
    def date(self) -> dt.date:
        return self.datetime.date()

    @date.setter
    @dispatch
    def date(self, value: dt.date) -> None:
        if isinstance(value, dt.datetime):
            raise NotImplementedError
        raise NotImplementedError

    @date.setter
    @dispatch
    def date(self, value: str) -> None:
        raise NotImplementedError

    @property
    def time(self) -> dt.time:
        return self.datetime.time()

    @time.setter
    @dispatch
    def time(self, value: dt.time) -> None:
        raise NotImplementedError

    @time.setter
    @dispatch
    def time(self, value: str) -> None:
        raise NotImplementedError

    @dispatch
    def __sub__(self, other: LikeMoment) -> Interval:
        """this - moment -> interval"""
        return Interval(self.datetime - upM(other))

    @dispatch
    def __sub__(self, other: LikeInterval) -> "Moment":
        """this - interval -> moment"""
        return Moment(self.datetime - upI(other))

    @dispatch
    def __rsub__(self, other: LikeMoment) -> Interval:
        """moment - this -> interval"""
        return Moment(other) - self

    @dispatch
    def __add__(self, other: LikeInterval) -> "Moment":
        """this + interval -> moment"""
        return Moment(self.datetime + upI(other))

    @dispatch
    def __lt__(self, other: LikeMoment) -> bool:
        """this < moment -> bool"""
        return self.datetime < upM(other)

    @dispatch
    def __le__(self, other: LikeMoment) -> bool:
        """this <= moment -> bool"""
        return self.datetime <= upM(other)

    @dispatch
    def __eq__(self, other: LikeMoment) -> bool:
        """this == moment -> bool"""
        return self.datetime == upM(other)

    @dispatch
    def __gt__(self, other: LikeMoment) -> bool:
        """this > moment -> bool"""
        return self.datetime > upM(other)

    @dispatch
    def __ge__(self, other: LikeMoment) -> bool:
        """this >= moment -> bool"""
        return self.datetime >= upM(other)

    def __str__(self) -> str:
        return convert(self.datetime, str)

    def __repr__(self) -> str:
        return str(self)

    def __float__(self) -> float:
        # FIXME: microseconds are not converted
        return time.mktime(self.datetime.timetuple())


IntervalType.deliver(Interval)
MomentType.deliver(Moment)

add_conversion_methods()
